//! Prymn messaging framework

use async_nats::SubscribeError;
use async_nats::{Client, ConnectError, ConnectOptions};
use thiserror::Error;
use tokio_stream::StreamExt;

pub use self::v1::OutgoingMessage;

mod v1 {
    use bytes::Bytes;

    pub(super) const PREFIX: &'static str = "agents.v1";

    #[derive(Debug)]
    pub enum OutgoingMessage {
        Health(String),
    }

    impl OutgoingMessage {
        pub fn subject(&self, sender: &str) -> String {
            match self {
                OutgoingMessage::Health(_) => format!("{}.{}.health", PREFIX, sender),
            }
        }

        pub fn into_payload(self) -> Bytes {
            match self {
                OutgoingMessage::Health(data) => Bytes::from(data.into_bytes()),
            }
        }
    }

    // #[derive(Debug)]
    // pub enum Subject {
    //     Command,
    // }

    // /// A system update message received when an OS update is requested
    // #[derive(Deserialize)]
    // struct SystemUpdate {
    //     /// `true` when a simulated update should occur
    //     simulate: bool,
    // }

    // /// An agent update message received when an agent binary update is requested
    // #[derive(Deserialize)]
    // struct AgentUpdate {
    //     /// `true` when a simulated update should occur
    //     simulate: bool,
    // }

    // /// An operating system program execution.
    // #[derive(Deserialize)]
    // struct Command<'a> {
    //     /// The user to which the child process will switch to
    //     user: &'a str,
    //     /// Program name or path
    //     program: &'a str,
    //     /// Program arguments
    //     args: Vec<&'a str>,
    // }

    // struct OpenTerminal {

    // }
}

#[derive(Debug, Error)]
pub enum MessagingError {
    #[error("failure connecting to message broker")]
    ConnectError(#[from] ConnectError),
    #[error("failure listening for messages from broker")]
    SubscribeError(#[from] SubscribeError),
}

#[derive(Debug, Clone)]
pub struct Agent {
    id: String,
    client: Client,
}

impl Agent {
    #[tracing::instrument]
    pub async fn connect(id: &str) -> Result<Self, MessagingError> {
        let client = ConnectOptions::new()
            .name(format!("Prymn Agent {id}"))
            .custom_inbox_prefix(format!("_INBOX_{id}"))
            .user_and_password("demo_agent".to_owned(), "demo_agent_password".to_owned())
            .connect("localhost")
            .await?;

        tracing::debug!("connected to NATS");
        let prefix = format!("{}.{}.*", v1::PREFIX, id);
        let mut routing_subscriber = client.subscribe(prefix.clone()).await?;
        tracing::debug!("root message routing initialized on {}", &prefix);

        tokio::spawn(async move {
            while let Some(data) = routing_subscriber.next().await {
                let command = data.subject.trim_start_matches(&prefix[..prefix.len() - 1]);
                dbg!(command);
            }
        });

        Ok(Self {
            id: id.to_owned(),
            client,
        })
    }

    #[tracing::instrument(skip(self))]
    pub async fn publish(&self, msg: OutgoingMessage) {
        let subj = msg.subject(&self.id);
        tracing::debug!(subject = &subj, "publish");
        if let Err(err) = self.client.publish(subj, msg.into_payload()).await {
            tracing::error!(cause = %err, "publish error");
        }
    }
}
